# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import numpy as np
from .utils import RandomParameterSearch, PARAMS
from .sparse_array import SparseArray
from .model import Ensemble
from multiprocessing import Pool
import EvoDAG as evodag
from EvoDAG import EvoDAG
import os
import gzip
import json
import pickle
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


def init_evodag(seed_args_X_y_test):
    seed, args, X, y, test = seed_args_X_y_test
    m = EvoDAG(seed=seed, **args).fit(X, y, test_set=test)
    return m.model()


def rs_evodag(args_X_y):
    args, X, y = args_X_y
    rs = RandomParameterSearch
    fit = []
    for seed in range(3):
        evo = EvoDAG(seed=seed,
                     **rs.process_params(args)).fit(X, y)
        fit.append(evo.model().fitness_vs)
    return fit, args


class CommandLine(object):
    def __init__(self):
        self.Xtest = None
        self.word2id = {}
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.training_set()
        self.init_params()
        self.model_file()
        self.test_set()
        self.optimize_parameters()
        self.cores()
        self.ensemble()
        self.output_file()
        self.version()

    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoDAG %s' % evodag.__version__)

    def output_file(self):
        self.parser.add_argument('-o', '--output-file',
                                 help='File to store the test set',
                                 dest='output_file',
                                 default=None,
                                 type=str)

    def ensemble(self):
        self.parser.add_argument('-n', '--ensemble-size',
                                 help='Ensemble size',
                                 dest='ensemble_size',
                                 default=1,
                                 type=int)

    def cores(self):
        self.parser.add_argument('-u', '--cpu-cores',
                                 help='Number of cores',
                                 dest='cpu_cores',
                                 default=1,
                                 type=int)

    def optimize_parameters(self):
        cdn = '''Optimize parameters sampling
        N points from the parameter space'''
        self.parser.add_argument('-r', '--optimize-parameters',
                                 dest='optimize_parameters',
                                 type=int, help=cdn)
        cdn = 'File to store the fitness of the configuration explored'
        self.parser.add_argument('--parameters',
                                 dest='parameters',
                                 type=str,
                                 help=cdn)

    def test_set(self):
        cdn = 'File containing the test set on csv.'
        self.parser.add_argument('-t', '--test_set',
                                 default=None, type=str,
                                 help=cdn)

    def model_file(self):
        pa = self.parser.add_argument
        pa('-m', '--model-file', type=str, dest='model_file',
           default=None,
           help='File name to either store or load the model')

    def init_params(self):
        pa = self.parser.add_argument
        pa('-c', '--classifier', dest='classifier',
           help='The task is either classification or regression',
           default=True, type=bool)
        pa('-e', '--early_stopping_rounds', dest='early_stopping_rounds',
           type=int,
           help='Early stopping rounds')
        pa('-p', '--popsize', dest='popsize',
           type=int, help='Population size')
        pa('-s', '--seed', dest='seed',
           default=0,
           type=int, help='Seed')
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)

    def training_set(self):
        cdn = 'File containing the training set on csv.'
        self.parser.add_argument('training_set',
                                 nargs='?',
                                 default=None,
                                 help=cdn)

    def parse_args(self):
        self.data = self.parser.parse_args()
        self.main()

    def convert(self, x):
        try:
            return float(x)
        except ValueError:
            if x not in self.word2id:
                self.word2id[x] = len(self.word2id)
            return self.word2id[x]

    def read_data(self, fname):
        with open(fname, 'r') as fpt:
            l = fpt.readlines()
        X = []
        for i in l:
            x = i.rstrip().lstrip()
            if len(x):
                X.append([self.convert(i) for i in x.split(',')])
        return X

    @staticmethod
    def _num_terms(a):
        if 'num_terms' in a:
            num_terms = a['num_terms']
        else:
            num_terms = len(a)
            if 'klass' in a:
                num_terms -= 1
        return num_terms

    def read_data_json(self, fname):
        import json
        X = None
        y = []
        with open(fname, 'r') as fpt:
            l = fpt.readlines()
        for row, d in enumerate(l):
            a = json.loads(d)
            if X is None:
                X = [list() for i in range(self._num_terms(a))]
            for k, v in a.items():
                try:
                    k = int(k)
                    X[k].append((row, self.convert(v)))
                except ValueError:
                    if k == 'klass' or k == 'y':
                        y.append(self.convert(v))
        num_rows = len(l)
        X = [SparseArray.init_index_data([i[0] for i in x],
                                         [i[1] for i in x],
                                         num_rows) for x in X]
        if len(y) == 0:
            y = None
        else:
            y = np.array(y)
        return X, y

    def read_training_set(self):
        if self.data.training_set is None:
            return
        if not self.data.json:
            d = np.array(self.read_data(self.data.training_set))
            self.X = d[:, :-1]
            self.y = d[:, -1]
            return True
        else:
            X, y = self.read_data_json(self.data.training_set)
            self.X = X
            self.y = y
            return True

    def read_test_set(self):
        if self.data.test_set is None:
            return False
        if not self.data.json:
            d = np.array(self.read_data(self.data.test_set))
            self.Xtest = d
            return True
        else:
            X, _ = self.read_data_json(self.data.test_set)
            self.Xtest = X
            return True

    def get_model_file(self):
        if self.data.model_file is None:
            a = self.data.training_set.split('.')[0]
            self.data.model_file = a + '.evodag.gz'
        return self.data.model_file

    def store_model(self, kw):
        if self.data.ensemble_size == 1:
            self.evo = EvoDAG(**kw).fit(self.X, self.y, test_set=self.Xtest)
            self.model = self.evo.model()
        else:
            seed = self.data.seed
            esize = self.data.ensemble_size
            args = [(x, kw, self.X, self.y, self.Xtest)
                    for x in range(seed, seed+esize)]
            if self.data.cpu_cores == 1:
                evo = [init_evodag(x) for x in tqdm(args, total=len(args))]
            else:
                p = Pool(self.data.cpu_cores)
                evo = [x for x in tqdm(p.imap_unordered(init_evodag, args),
                                       total=len(args))]
                p.close()
            self.model = Ensemble(evo)
        model_file = self.get_model_file()
        with gzip.open(model_file, 'w') as fpt:
            pickle.dump(self.model, fpt)
            pickle.dump(self.word2id, fpt)

    def evolve(self, kw):
        if os.path.isfile(self.get_model_file()):
            with gzip.open(self.get_model_file(), 'r') as fpt:
                self.model = pickle.load(fpt)
                self.word2id = pickle.load(fpt)
                return
        if self.data.optimize_parameters is not None:
            if len(kw):
                params = PARAMS.copy()
                for k, v in kw.items():
                    if k in params and v is not None:
                        params[k] = [v]
            parameters = self.data.parameters
            if parameters is not None and os.path.isfile(parameters):
                with gzip.open(self.data.parameters, 'r') as fpt:
                    res = pickle.load(fpt)
            else:
                npoints = self.data.optimize_parameters
                rs = RandomParameterSearch(params=params,
                                           seed=self.data.seed,
                                           npoints=npoints)
                if self.data.cpu_cores == 1:
                    res = [rs_evodag((args, self.X, self.y))
                           for args in tqdm(rs, total=rs._npoints)]
                else:
                    p = Pool(self.data.cpu_cores)
                    args = [(args, self.X, self.y) for args in rs]
                    res = [x for x in tqdm(p.imap_unordered(rs_evodag, args),
                                           total=len(args))]
                    p.close()
                res.sort(key=lambda x: np.median(x[0]), reverse=True)
                if self.data.parameters is not None:
                    with gzip.open(self.data.parameters, 'w') as fpt:
                        pickle.dump(res, fpt)
            kw = RandomParameterSearch.process_params(res[0][1])
        self.store_model(kw)

    def get_output_file(self):
        if self.data.output_file is None:
            self.data.output_file = self.data.test_set + '.evodag.csv'
            # if self.data.json:
            #     self.data.output_file += '.json'
            # else:
            #     self.data.output_file += '.csv'
        return self.data.output_file

    def id2word(self, x):
        if not self.data.classifier:
            return x
        if len(self.word2id) == 0:
            return x
        i2w = dict([(i[1], i[0]) for i in self.word2id.items()])
        return [i2w[int(i)] for i in x]

    def main(self):
        self.read_training_set()
        test_set = self.read_test_set()
        kw = {}
        for k, v in EvoDAG().get_params().items():
            if hasattr(self.data, k):
                kw[k] = getattr(self.data, k)
        self.evolve(kw)
        if test_set:
            hy = self.id2word(self.model.predict(self.Xtest))
            with open(self.get_output_file(), 'w') as fpt:
                fpt.write('\n'.join(map(str, hy)))


class CommandLineParams(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.word2id = {}
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.training_set()
        self.init_params()
        self.optimize_parameters()
        self.cores()
        self.version()

    def optimize_parameters(self):
        cdn = '''Optimize parameters sampling
        N (734 by default) points from the parameter space'''
        self.parser.add_argument('-r', '--optimize-parameters',
                                 dest='optimize_parameters',
                                 default=734,
                                 type=int, help=cdn)
        cdn = 'File to store the fitness of the parameters explored'
        self.parser.add_argument('-P', '--parameters',
                                 dest='parameters',
                                 type=str,
                                 help=cdn)

    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoDAG %s' % evodag.__version__)

    def evolve(self, kw):
        if len(kw):
            params = PARAMS.copy()
            for k, v in kw.items():
                if k in params and v is not None:
                    params[k] = [v]
        parameters = self.data.parameters
        if parameters is None:
            parameters = self.data.training_set + '.EvoDAGparams'
        npoints = self.data.optimize_parameters
        rs = RandomParameterSearch(params=params,
                                   seed=self.data.seed,
                                   npoints=npoints)
        if self.data.cpu_cores == 1:
            res = [rs_evodag((args, self.X, self.y))
                   for args in tqdm(rs, total=rs._npoints)]
        else:
            p = Pool(self.data.cpu_cores)
            args = [(args, self.X, self.y) for args in rs]
            res = [x for x in tqdm(p.imap_unordered(rs_evodag, args),
                                   total=len(args))]
            p.close()
        [x[1].update(dict(fitness=x[0])) for x in res]
        res = [x[1] for x in res]
        [x.update(kw) for x in res]
        res.sort(key=lambda x: np.median(x['fitness']), reverse=True)
        if parameters.endswith('.gz'):
            with gzip.open(parameters, 'wb') as fpt:
                try:
                    fpt.write(bytes(json.dumps(res, sort_keys=True), encoding='utf-8'))
                except TypeError:
                    fpt.write(json.dumps(res, sort_keys=True))
        else:
            with open(parameters, 'w') as fpt:
                fpt.write(json.dumps(res, sort_keys=True))

    def main(self):
        self.read_training_set()
        kw = {}
        for k, v in EvoDAG().get_params().items():
            if hasattr(self.data, k) and getattr(self.data, k) is not None:
                kw[k] = getattr(self.data, k)
        self.evolve(kw)


class CommandLineTrain(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.word2id = {}
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.training_set()
        self.parameters()
        self.model()
        self.cores()
        self.ensemble()
        self.test_set()
        self.version()

    def parameters(self):
        cdn = 'File containing a list of parameters explored,\
        the first one being the best'
        self.parser.add_argument('-P', '--parameters',
                                 dest='parameters',
                                 type=str,
                                 help=cdn)

    def model(self):
        cdn = 'File to store EvoDAG model'
        pa = self.parser.add_argument
        pa('-m', '--model',
           dest='model_file',
           type=str,
           help=cdn)
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)

    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoDAG %s' % evodag.__version__)

    def main(self):
        self.read_training_set()
        self.read_test_set()
        parameters = self.data.parameters
        if parameters.endswith('.gz'):
            func = gzip.open
        else:
            func = open
        with func(parameters, 'rb') as fpt:
            try:
                d = fpt.read()
                res = json.loads(str(d, encoding='utf-8'))
            except TypeError:
                res = json.loads(d)
        try:
            kw = res[0]
        except KeyError:
            kw = res
        kw = RandomParameterSearch.process_params(kw)
        if 'seed' in kw:
            self.data.seed = kw['seed']
            del kw['seed']
        self.store_model(kw)


class CommandLinePredict(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.word2id = {}
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.model()
        self.test_set()
        self.output_file()
        self.version()

    def test_set(self):
        cdn = 'File containing the test set on csv.'
        self.parser.add_argument('test_set',
                                 default=None,
                                 help=cdn)

    def model(self):
        cdn = 'EvoDAG model'
        pa = self.parser.add_argument
        pa('-m', '--model',
           dest='model_file',
           type=str,
           help=cdn)
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)

    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoDAG %s' % evodag.__version__)

    def main(self):
        self.read_test_set()
        model_file = self.get_model_file()
        with gzip.open(model_file, 'r') as fpt:
            m = pickle.load(fpt)
            self.word2id = pickle.load(fpt)
        self.data.classifier = m.classifier
        hy = self.id2word(m.predict(self.Xtest))
        with open(self.get_output_file(), 'w') as fpt:
            fpt.write('\n'.join(map(str, hy)))


def main():
    "Command line main"
    c = CommandLine()
    c.parse_args()


def params():
    "EvoDAG-params command line"
    c = CommandLineParams()
    c.parse_args()


def train():
    "EvoDAG-params command line"
    c = CommandLineTrain()
    c.parse_args()


def predict():
    "EvoDAG-params command line"
    c = CommandLinePredict()
    c.parse_args()




# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
from EvoDAG.command_line import CommandLine, main
from test_root import X, cl


def training_set():
    import tempfile
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(v)
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    return fname


def test_command_line():
    fname = training_set()
    sys.argv = ['EvoDAG', '-s', '1', '-e', '10', '-p', '100', fname]
    c = CommandLine()
    c.parse_args()
    assert c.data.training_set == fname
    os.unlink(fname)
    os.unlink(c.data.model_file)
    assert c.evo._early_stopping_rounds == 10
    assert c.evo._classifier
    assert c.evo._popsize == 100
    assert c.evo._seed == 1


def test_main():
    fname = training_set()
    sys.argv = ['EvoDAG', '-m', 'temp.evodag.gz',
                '-e', '10', '-p', '100', fname, '-t', fname]
    main()
    os.unlink(fname)
    os.unlink('temp.evodag.gz')
    os.unlink(fname + '.evodag.csv')


def test_optimize_parameters():
    import os
    fname = training_set()
    sys.argv = ['EvoDAG', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    c = CommandLine()
    c.parse_args()
    os.unlink(fname)
    os.unlink(c.data.model_file)
    assert c.evo._popsize == 10
    assert os.path.isfile('cache.evodag.gz')
    os.unlink('cache.evodag.gz')


def test_previous_model():
    import gzip
    import pickle
    fname = training_set()
    if os.path.isfile('model.evodag.gz'):
        os.unlink('model.evodag.gz')
    sys.argv = ['EvoDAG', '-p10', '-e2', '-m', 'model.evodag.gz', fname]
    c = CommandLine()
    c.parse_args()
    with gzip.open(c.data.model_file, 'w') as fpt:
        pickle.dump([], fpt)
        pickle.dump([], fpt)
    c = CommandLine()
    c.parse_args()
    os.unlink(fname)
    os.unlink(c.data.model_file)
    assert isinstance(c.model, list) and len(c.model) == 0


def test_cpu_cores():
    import os
    fname = training_set()
    sys.argv = ['EvoDAG', '-u2', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    c = CommandLine()
    c.parse_args()
    assert c.evo._popsize == 10
    assert os.path.isfile('cache.evodag.gz')
    os.unlink(c.data.model_file)
    c = CommandLine()
    c.parse_args()
    os.unlink('cache.evodag.gz')
    os.unlink(fname)
    os.unlink(c.data.model_file)
    

def test_ensemble_size():
    import os
    from EvoDAG.model import Ensemble
    fname = training_set()
    sys.argv = ['EvoDAG', '-u2', '-n2', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    c = CommandLine()
    c.parse_args()
    os.unlink(fname)
    os.unlink(c.data.model_file)
    assert os.path.isfile('cache.evodag.gz')
    assert isinstance(c.model, Ensemble)
    os.unlink('cache.evodag.gz')

    
def test_word2id():
    import tempfile
    fname = tempfile.mktemp()
    id2word = dict([[0, 'a'], [1, 'b'], [2, 'c']])
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(id2word[v])
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    sys.argv = ['EvoDAG', '-e2', '-p10', fname]
    c = CommandLine()
    c.parse_args()
    os.unlink(fname)


def test_output_file():
    fname = training_set()
    print(fname)
    sys.argv = ['EvoDAG', '-e2', '-p10',
                '-o', 'output.evodag.csv', '-t', fname,
                fname]
    c = CommandLine()
    c.parse_args()
    os.unlink(fname)
    print(os.path.isfile('output.evodag.csv'))
    assert os.path.isfile('output.evodag.csv')
    os.unlink('output.evodag.csv')
    

def test_id2word():
    import tempfile
    fname = tempfile.mktemp()
    id2word = dict([[0, 'a'], [1, 'b'], [2, 'c']])
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(id2word[v])
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    sys.argv = ['EvoDAG', '-e2', '-p10', fname, '-t',
                fname, '-o', 'output.txt']
    c = CommandLine()
    c.parse_args()
    with open('output.txt', 'r') as fpt:
        a = fpt.readlines()
    for i in a:
        assert i.rstrip() in ['a', 'b', 'c']
    os.unlink(fname)
    os.unlink('output.txt')
    

def test_json():
    import tempfile
    import json
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, y in zip(X, cl):
            a = {k: v for k, v in enumerate(x)}
            a['klass'] = int(y)
            a['num_terms'] = len(x)
            fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-m', 'temp.evodag.gz', '--json',
                '-e', '10', '-p', '100', fname, '-ooutput.evodag', '-t', fname]
    main()
    os.unlink(fname)
    os.unlink('temp.evodag.gz')
    print(open('output.evodag').read())
    os.unlink('output.evodag')


def test_params():
    import os
    import gzip
    from EvoDAG.command_line import params
    import json
    fname = training_set()
    sys.argv = ['EvoDAG', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    params()
    os.unlink(fname)
    assert os.path.isfile('cache.evodag.gz')
    with gzip.open('cache.evodag.gz', 'rb') as fpt:
        try:
            d = fpt.read()
            a = json.loads(str(d, encoding='utf-8'))
        except TypeError:
            a = json.loads(d)
    os.unlink('cache.evodag.gz')
    assert len(a) == len([x for x in a if 'fitness' in x])
    print(a)


def test_train():
    import os
    from EvoDAG.command_line import params, train
    fname = training_set()
    sys.argv = ['EvoDAG', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag.gz',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    os.unlink(fname)
    os.unlink('cache.evodag.gz')
    assert os.path.isfile('model.evodag')
    os.unlink('model.evodag')


def test_predict():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = training_set()
    sys.argv = ['EvoDAG', '--parameters',
                'cache.evodag.gz', '-p10', '-e2', '-r', '2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag.gz',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag.gz')
    os.unlink('model.evodag')
    assert os.path.isfile('output.evodag')
    os.unlink('output.evodag')
    



from test_root import X, cl
import sys
from EvoDAG.command_line import main
import tempfile
import json
import os

fname = tempfile.mktemp()
with open(fname, 'w') as fpt:
    for x, y in zip(X, cl):
        a = {k: v for k, v in enumerate(x)}
        a['klass'] = int(y)
        a['num_terms'] = len(x)
        fpt.write(json.dumps(a) + '\n')
sys.argv = ['EvoDAG', '-m', 'temp.evodag.gz', '--json', '-s2',
            '-e', '10', '-p', '100', fname]  # , '-ooutput.evodag', '-t', fname]
print("Empezando el main")
main()
print("Terminando el main")
os.unlink(fname)
os.unlink('temp.evodag.gz')
print(open('output.evodag').read())
os.unlink('output.evodag')



# Copyright 2013 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#cython: profile=True
#cython: nonecheck=False


from array import array as p_array
# cimport numpy as npc
cimport cython
import numpy as np
import types
from cpython.mem cimport PyMem_Malloc, PyMem_Free
cdef extern from "numpy/npy_math.h":
    bint npy_isinf(double)
    bint npy_isnan(double)
    long double INFINITY "NPY_INFINITY"
    long double PI "NPY_PI"


cpdef rebuild(data, index, size):
    cdef SparseArray r = SparseArray()
    r.init(len(data))
    r.set_size(size)
    r.set_data_index(data, index)
    return r


@cython.freelist(512)
cdef class SparseArray:
    def __cinit__(self):
        self._size = 0
        self._nele = 0
        self._usemem = 0

    cpdef int size(self):
        return self._size

    cpdef set_size(self, int s):
        self._size = s

    cpdef int nele(self):
        return self._nele

    cpdef set_nele(self, int s):
        self._nele = s

    cpdef init_ptr(self, array.array[int] index,
                   array.array[double] data):
        self._indexC = index.data.as_ints
        self._dataC = data.data.as_doubles

    cpdef init(self, int nele):
        if nele == 0:
            self._usemem = 1
            return
        if self._usemem == 0:
            # print "Poniendo memoria", self
            self._usemem = 1
            self._dataC = <double*> PyMem_Malloc(nele * sizeof(double))
            if not self._dataC:
                raise MemoryError("dataC")
            self._indexC = <int *> PyMem_Malloc(nele * sizeof(int))
            if not self._indexC:
                raise MemoryError("indexC")
            self.set_nele(nele)
        else:
            print "Error init", self

    def __dealloc__(self):
        if self._usemem and self.nele() == 0:
            return
        if self._usemem:
            # print "Borrando", self
            self._usemem = 0
            PyMem_Free(self._dataC)
            PyMem_Free(self._indexC)
            # self._dataC = null
            # self._indexC = null
        else:
            print "Error dealloc", self
            
    cpdef int nunion(self, SparseArray other):
        cdef int a=0, b=0, c=0
        cdef int anele = self.nele(), bnele=other.nele()
        if self.size() == anele or self.size() == bnele:
            return self.size()
        while (a < anele) and (b < bnele):
            if self._indexC[a] == other._indexC[b]:
                a += 1
                b += 1
            elif self._indexC[a] < other._indexC[b]:
                a += 1
            else:
                b += 1
            c += 1
        if a == anele:
            return c + bnele - b
        else:
            return c + anele - a

    cpdef int nintersection(self, SparseArray other):
        cdef int a=0, b=0, c=0, *indexC = self._indexC, *oindexC = other._indexC
        cdef int anele = self.nele(), bnele=other.nele()
        if self.size() == anele or self.size() == bnele:
            if anele < bnele:
                return anele
            return bnele
        while (a < anele) and (b < bnele):
            if indexC[a] == oindexC[b]:
                a += 1
                b += 1
                c += 1
            elif indexC[a] < oindexC[b]:
                a += 1
            else:
                b += 1                
        return c

    def __add__(self, other):
        if isinstance(other, SparseArray):
            return self.add(other)
        return self.add2(other)

    cpdef SparseArray add(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nunion(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                res._dataC[c] = self._dataC[c] + other._dataC[c]
                res._indexC[c] = c
            return res
        for c in range(rnele):
            if a >= anele:
                index = other._indexC[b]
                r = other._dataC[b]
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    r = self._dataC[a] + other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    a += 1
                else:
                    r = other._dataC[b]
                    index = other._indexC[b]
                    b += 1
            res._dataC[c] = r
            res._indexC[c] = index
        return res

    cpdef SparseArray add2(self, double other):
        cdef SparseArray res = self.empty(self.size(), self.size())
        cdef int i
        for i in range(self.size()):
            res._indexC[i] = i
            res._dataC[i] = other
        for i in range(self.nele()):
            res._dataC[self._indexC[i]] = self._dataC[i] + other
        return res

    def __sub__(self, other):
        if isinstance(other, SparseArray):
            return self.sub(other)
        return self.sub2(other)
        
    cpdef SparseArray sub(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nunion(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                res._dataC[c] = self._dataC[c] - other._dataC[c]
                res._indexC[c] = c
            return res        
        for c in range(rnele):
            if a >= anele:
                index = other._indexC[b]
                r = - other._dataC[b]
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    r = self._dataC[a] - other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    a += 1
                else:
                    r = - other._dataC[b]
                    index = other._indexC[b]
                    b += 1
            res._dataC[c] = r
            res._indexC[c] = index
        return res

    cpdef SparseArray sub2(self, double other):
        cdef SparseArray res = self.empty(self.size(), self.size())
        cdef int i
        for i in range(self.size()):
            res._indexC[i] = i
            res._dataC[i] = - other
        for i in range(self.nele()):
            res._dataC[self._indexC[i]] = self._dataC[i] - other
        return res
        
    def __mul__(self, other):
        if isinstance(other, SparseArray):
            return self.mul(other)
        if other == 1:
            return self
        return self.mul2(other)

    cpdef SparseArray mul(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nintersection(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                res._dataC[c] = self._dataC[c] * other._dataC[c]
                res._indexC[c] = c
            return res        
        while c < rnele:
            if a >= anele:
                b += 1
            elif b >= bnele:
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    res._dataC[c] = self._dataC[a] * other._dataC[b]
                    res._indexC[c] = index
                    a += 1; b += 1; c += 1
                elif index < other._indexC[b]:
                    a += 1
                else:
                    b += 1
        return res

    cpdef SparseArray mul2(self, double other):
        cdef SparseArray res = self.empty(self.nele(), self.size())
        cdef int i
        for i in range(self.nele()):
            res._indexC[i] = self._indexC[i]
            res._dataC[i] = self._dataC[i] * other
        return res
                    
    def __div__(self, other):
        if isinstance(other, SparseArray):
            return self.div(other)
        return self.div2(other)

    def __truediv__(self, other):
        if isinstance(other, SparseArray):
            return self.div(other)
        return self.div2(other)

    @cython.cdivision(True)
    cpdef SparseArray div(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nintersection(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                res._dataC[c] = self._dataC[c] / other._dataC[c]
                res._indexC[c] = c
            return res        
        while c < rnele:
            if a >= anele:
                b += 1
            elif b >= bnele:
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    res._dataC[c] = self._dataC[a] / other._dataC[b]
                    res._indexC[c] = index
                    a += 1; b += 1; c += 1
                elif index < other._indexC[b]:
                    a += 1
                else:
                    b += 1
        return res

    @cython.cdivision(True)    
    cpdef SparseArray div2(self, double other):
        cdef SparseArray res = self.empty(self.nele(), self.size())
        cdef int i
        for i in range(self.nele()):
            res._indexC[i] = self._indexC[i]
            res._dataC[i] = self._dataC[i] / other
        return res
        
    cpdef double sum(self):
        cdef double res=0, *data = self._dataC
        cdef int i
        for i in xrange(self._nele):
            res += data[i]
        return res

    cpdef double mean(self):
        cdef double res=0
        res = self.sum()    
        return res / self.size()

    @cython.boundscheck(False)
    @cython.nonecheck(False)    
    def var_per_cl(self, list X, list mu,
                   array.array[double] kfreq):
        cdef int i, k=0, ncl = len(kfreq), j, xnele, ynele=self.nele()
        cdef SparseArray x
        cdef list m=[]
        cdef double epsilon = 1e-9
        cdef array.array[double] var, mu_x
        for x, mu_x in zip(X, mu):
            var = array.array('d', [0] * ncl)
            k = 0
            i = 0
            xnele = x.nele()
            while i < xnele:
                if x._indexC[i] == self._indexC[k]:
                    j = int(self._dataC[k])
                    var[j] += math.pow(x._dataC[i] - mu_x[j], 2)
                    k += 1
                    i += 1
                elif self._indexC[k] < x._indexC[i]:
                    if k < ynele:
                        k += 1
                    else:
                        var[0] += math.pow(x._dataC[i] - mu_x[0], 2)
                        i += 1
                else:
                    var[0] += math.pow(x._dataC[i] - mu_x[0], 2)
                    i += 1
            for i in range(ncl):
                var[i] = var[i] / kfreq[i] + epsilon
            m.append(var)
        return m

    @cython.boundscheck(False)
    @cython.nonecheck(False)    
    def mean_per_cl(self, list X, array.array[double] kfreq):
        cdef int i, k=0, ncl = len(kfreq), j, xnele, ynele=self.nele()
        cdef SparseArray x
        cdef list m=[]
        cdef array.array[double] mu
        for x in X:
            mu = array.array('d', [0] * ncl)
            k = 0
            i = 0
            xnele = x.nele()
            while i < xnele:
                if x._indexC[i] == self._indexC[k]:
                    j = int(self._dataC[k])
                    mu[j] += x._dataC[i]
                    k += 1
                    i += 1
                elif self._indexC[k] < x._indexC[i]:
                    if k < ynele:
                        k += 1
                    else:
                        mu[0] += x._dataC[i]
                        i += 1
                else:
                    mu[0] += x._dataC[i]
                    i += 1
            for i in range(ncl):
                mu[i] = mu[i] / kfreq[i]
            m.append(mu)
        return m

    @cython.boundscheck(False)
    @cython.nonecheck(False)        
    def class_freq(self, int ncl):
        cdef int i
        cdef list f
        cdef array.array[double] mu = array.array('d', [0] * ncl)
        for i in range(self.nele()):
            mu[int(self._dataC[i])] += 1.0
        mu[0] = self.size() - sum(mu[1:])
        return mu    
        
    cpdef double std(self):
        cdef SparseArray res = self.sub2(self.mean())
        res = res.sq()
        return math.sqrt(res.sum() / res.size())

    cpdef SparseArray fabs(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        for i in xrange(self.nele()):
            res._dataC[i] = math.fabs(self._dataC[i])
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray exp(self):
        cdef SparseArray res = self.empty(self.size(), self.size())
        cdef int i, j=0, nele=self.nele()
        for i in xrange(self.size()):
            if j < nele and self._indexC[j] == i:
                res._dataC[i] = math.exp(self._dataC[j])
                j += 1
            else:
                res._dataC[i] = 1
            res._indexC[i] = i
        return res
        
    cpdef SparseArray sin(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        for i in xrange(self.nele()):
            res._dataC[i] = math.sin(self._dataC[i])
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray cos(self):
        cdef SparseArray res = self.empty(self.size(), self.size())
        cdef int i, j=0, nele=self.nele()
        for i in xrange(self.size()):
            if j < nele and self._indexC[j] == i:
                res._dataC[i] = math.cos(self._dataC[j])
                j += 1
            else:
                res._dataC[i] = 1
            res._indexC[i] = i
        return res
        
    cpdef SparseArray ln(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            r = self._dataC[i]
            res._dataC[i] = math.log(math.fabs(r))
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray sq(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            r = self._dataC[i]
            res._dataC[i] = r * r
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray sqrt(self):
        cdef SparseArray res = self.empty(self.nele(), self.size())
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            res._dataC[i] = math.sqrt(self._dataC[i])
            res._indexC[i] = self._indexC[i]
        return res

    @cython.cdivision(True)
    cpdef SparseArray sigmoid(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            res._dataC[i] = 1 / (1 + math.exp((-self._dataC[i] + 1) * 30))
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray if_func(self, SparseArray y, SparseArray z):
        cdef SparseArray s = self.sigmoid()
        cdef SparseArray sy, sz
        cdef SparseArray r
        cdef SparseArray res
        sy = s.mul(y)
        sz = s.mul(z)
        r = sy.sub(sz)
        res = r.add(z)
        return res

    cpdef SparseArray sign(self):
        cdef SparseArray res = self.empty(self.nele(), self.size())
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            r = self._dataC[i]
            res._dataC[i] = 0
            if r > 0:
                res._dataC[i] = 1
            elif r < 0:
                res._dataC[i] = -1
            res._indexC[i] = self._indexC[i]
        return res

    cpdef SparseArray min(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nunion(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                if self._dataC[c] < other._dataC[c]:
                    res._dataC[c] = self._dataC[c]
                else:
                    res._dataC[c] = other._dataC[c]
                res._indexC[c] = c
            return res
        for c in range(rnele):
            if a >= anele:
                index = other._indexC[b]
                r = other._dataC[b]
                if r > 0:
                    r = 0
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                if r > 0:
                    r = 0                
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    if self._dataC[a] < other._dataC[b]:
                        r = self._dataC[a]
                    else:
                        r = other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    if r > 0:
                        r = 0
                    a += 1
                else:
                    r = other._dataC[b]
                    if r > 0:
                        r = 0
                    index = other._indexC[b]
                    b += 1
            res._dataC[c] = r
            res._indexC[c] = index
        return res

    cpdef SparseArray max(self, SparseArray other):
        cdef SparseArray res = self.empty(self.nunion(other), self.size())
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self.nele(), bnele=other.nele(), rnele=res.nele()
        cdef double r
        if anele == rnele and bnele == rnele and res.size() == rnele:
            for c in range(rnele):
                if self._dataC[c] > other._dataC[c]:
                    res._dataC[c] = self._dataC[c]
                else:
                    res._dataC[c] = other._dataC[c]
                res._indexC[c] = c
            return res
        for c in range(rnele):
            if a >= anele:
                index = other._indexC[b]
                r = other._dataC[b]
                if r < 0:
                    r = 0
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                if r < 0:
                    r = 0                
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    if self._dataC[a] > other._dataC[b]:
                        r = self._dataC[a]
                    else:
                        r = other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    if r < 0:
                        r = 0
                    a += 1
                else:
                    r = other._dataC[b]
                    if r < 0:
                        r = 0
                    index = other._indexC[b]
                    b += 1
            res._dataC[c] = r
            res._indexC[c] = index
        return res

    cpdef SparseArray boundaries(self, float lower=-1, float upper=1):
        cdef SparseArray res = self.empty(self.nele(), self.size())
        cdef int i
        cdef double r
        for i in xrange(self.nele()):
            r = self._dataC[i]
            if r > upper:
                res._dataC[i] = upper
            elif r < lower:
                res._dataC[i] = lower
            else:
                res._dataC[i] = r                
            res._indexC[i] = self._indexC[i]
        return res
    
    cpdef double SAE(self, SparseArray other):
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self._nele, bnele=other._nele
        cdef double r, res=0
        cdef SparseArray last
        while True:
            if a >= anele and b >= bnele:
                break
            elif a >= anele:
                index = other._indexC[b]
                r = - other._dataC[b]
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    r = self._dataC[a] - other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    a += 1
                else:
                    r = - other._dataC[b]
                    index = other._indexC[b]
                    b += 1
            res +=  math.fabs(r)
        if npy_isnan(res):
            return INFINITY
        return res
            
    cpdef double SSE(self, SparseArray other):
        cdef int a=0, b=0, index=0, c=0
        cdef int anele = self._nele, bnele=other._nele
        cdef double r, res=0
        cdef SparseArray last
        while True:
            if a >= anele and b >= bnele:
                break
            elif a >= anele:
                index = other._indexC[b]
                r = - other._dataC[b]
                b += 1
            elif b >= bnele:
                index = self._indexC[a]
                r = self._dataC[a]
                a += 1
            else:
                index = self._indexC[a]
                if index == other._indexC[b]:
                    r = self._dataC[a] - other._dataC[b]
                    a += 1; b += 1
                elif index < other._indexC[b]:
                    r = self._dataC[a]
                    a += 1
                else:
                    r = - other._dataC[b]
                    index = other._indexC[b]
                    b += 1
            res +=  r * r
        if npy_isnan(res):
            return INFINITY
        return res

    cpdef double pearsonr(self, SparseArray other):
        cdef double mx, my, up
        mx = self.sum() / self.size()
        my = other.sum() / other.size()
        up = ((self - mx) * (other - my)).sum()
        mx = math.sqrt((self - mx).sq().sum())
        my = math.sqrt((other - my).sq().sum())
        if mx == 0 or my == 0:
            return 1
        return up / (mx * my)
            
    cpdef bint isfinite(self):
        cdef int i
        cdef double r    
        for i in range(self.nele()):
            r = self._dataC[i]
            if npy_isnan(r) or npy_isinf(r):
                return 0
        return 1

    cdef SparseArray select(self, npc.ndarray[long, ndim=1] index):
        cdef long *indexC = <long *>index.data
        cdef list data = [], index2 = []
        cdef int anele=self.nele(), bnele=index.shape[0]
        cdef int a=0, b=0, c=0, i=0
        cdef SparseArray res
        for i in range(index.shape[0]-1):
            if index[i] > index[i+1]:
                raise NotImplementedError("The index must be in order")
        while (a < anele) and (b < bnele):
            if self._indexC[a] == indexC[b]:
                data.append(self._dataC[a])
                index2.append(c)
                b += 1
                c += 1
                if (b >= bnele) or self._indexC[a] != indexC[b]:
                    a += 1
            elif self._indexC[a] < indexC[b]:
                a += 1
            else:
                b += 1
                c += 1
        res = self.empty(len(data), index.shape[0])
        res.set_data_index(data, index2)        
        return res
                
    def __getitem__(self, value):
        cdef int i, init=-1, cnt=0
        cdef SparseArray res
        if isinstance(value, np.ndarray):
            return self.select(value)
        if not isinstance(value, slice):
            raise NotImplementedError("Not implemented yet %s" %(type(value)))
        start = value.start if value.start is not None else 0
        stop = value.stop if value.stop is not None else self.size()
        if stop > self.size():
            stop = self.size()
        for i in range(self.nele()):
            if self._indexC[i] >= start and init == -1:
                init = i
            if init > -1 and self._indexC[i] < stop:
                cnt += 1
            if self._indexC[i] >= stop:
                break
        res = self.empty(cnt, stop - start)
        for i in range(init, init+cnt):
            res._indexC[i - init] = self._indexC[i] - start
            res._dataC[i - init] = self._dataC[i]
        return res

    def concatenate(self, SparseArray dos):
        cdef SparseArray res = self.empty(self.nele() + dos.nele(),
                                          self.size() + dos.size())
        cdef int i, j=0, size=self.size()
        for i in range(self.nele()):
            res._indexC[i] = self._indexC[i]
            res._dataC[i] = self._dataC[i]
        for i in range(self.nele(), res.nele()):
            res._indexC[i] = dos._indexC[j] + size
            res._dataC[i] = dos._dataC[j] 
            j += 1
        return res

    def tonparray(self):
        import numpy as np
        cdef npc.ndarray[double, ndim=1] res = np.zeros(self.size())
        cdef double * resC = <double *>res.data
        cdef int i, ele
        for i in range(self.nele()):
            ele = self._indexC[i]
            resC[ele] = self._dataC[i]
        return res

    def get_data(self):
        cdef list d=[]
        cdef int i
        for i in range(self.nele()):
            d.append(self._dataC[i])
        return d

    def get_index(self):
        cdef list d=[]
        cdef int i
        for i in range(self.nele()):
            d.append(self._indexC[i])
        return d    
        
    def tolist(self):
        cdef int i, ele
        lst = [0] * self.size()
        for i in range(self.nele()):
            ele = self._indexC[i]
            lst[ele] = self._dataC[i]
        return lst

    def set_data_index(self, data, index):
        cdef int c=0
        for d, i in zip(data, index):
            self._dataC[c] = d
            self._indexC[c] = i
            c += 1

    def print_data(self):
        cdef int i
        for i in range(self.nele()):
            print self._dataC[i]

    @classmethod
    def init_index_data(cls, list index, list data, int _size):
        self = cls()
        self.init(len(data))
        self.set_data_index(data, index)    
        self.set_size(_size)
        return self
        
    @classmethod
    def fromlist(cls, iter, bint force_finite=False):
        self = cls()
        data = []
        index = []
        k = -1
        for k, v in enumerate(iter):
            if force_finite and (npy_isnan(v) or npy_isinf(v)):
                continue
            if v == 0:
                continue
            data.append(v)
            index.append(k)
        self.init(len(data))
        self.set_data_index(data, index)    
        self.set_size(k + 1)
        return self

    cpdef SparseArray empty(self, int nele, int size=-1):
        cdef SparseArray res = SparseArray()
        res.init(nele)
        if size == -1:
            res.set_size(nele)
        else:
            res.set_size(size)
        return res

    cpdef SparseArray copy(self):
        cdef SparseArray res = self.empty(self.nele(), self._size)
        cdef int i
        for i in range(self.nele()):
            res._indexC[i] = self._indexC[i]
            res._dataC[i] = self._dataC[i]
        return res
            
    cpdef SparseArray constant(self, double v, int size=-1):
        cdef int i
        cdef SparseArray res = SparseArray()
        if size == -1:
            size = self.size()
        res.set_size(size)        
        if v == 0:
            res.init(0)
        else:
            res.init(size)
        for i in range(res.nele()):
            res._dataC[i] = v
            res._indexC[i] = i
        return res

    @cython.boundscheck(False)
    @cython.nonecheck(False)        
    def BER(self, SparseArray yh, array.array[double] class_freq):
        cdef array.array[double] err = array.array('d', [0] * len(class_freq))
        cdef int i=0, j=0, k=0, ynele=self.nele(), yhnele=yh.nele()
        cdef int c1=0, c2=0
        cdef double res=0
        for k in range(self.size()):
            if self._indexC[i] == k:
                c1 = int(self._dataC[i])
                if i < ynele - 1:
                    i += 1
            else:
                c1 = 0
            if j < yhnele and yh._indexC[j] == k:
                c2 = int(yh._dataC[j])
                j += 1
            else:
                c2 = 0
            if c1 != c2:
                err[c1] += 1
        for i in range(len(class_freq)):
            res += err[i] / class_freq[i]
        return res / len(class_freq) * 100.
        
    @staticmethod
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    def distance(list X, list ev, npc.ndarray[double, ndim=2] output):
        cdef SparseArray x
        cdef SparseArray y
        cdef double *data = <double *> output.data
        cdef int c = 0, i=0, j=0, len_X = len(X)
        for i in range(len(ev)):
            x = ev[i]
            for j in range(len_X):
                y = X[j]
                data[c] = x.SAE(y)
                c += 1

    def __reduce__(self):
        cdef list data = []
        cdef list index = []
        cdef int i
        for i in range(self.nele()):
            data.append(self._dataC[i])
            index.append(self._indexC[i])
        return (rebuild, (data, index, self.size()))



# Copyright 2013 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy
from setuptools import setup
from setuptools import Extension
from Cython.Distutils import build_ext
from distutils import sysconfig
from distutils.command.clean import clean
import os
from os.path import join


class Clean(clean):
    def run(self):
        clean.run(self)
        if self.all:
            for dir, dirs, files in os.walk('EvoDAG'):
                abspath = os.path.abspath(dir)
                for f in files:
                    ext = f.split('.')[-1]
                    if ext in ['c', 'so']:
                        os.unlink(os.path.join(abspath, f))


# -mno-fused-madd

lst = ['CFLAGS', 'CONFIG_ARGS', 'LIBTOOL', 'PY_CFLAGS']
for k, v in zip(lst, sysconfig.get_config_vars(*lst)):
    if v is None:
        continue
    v = v.replace('-mno-fused-madd', '')
    os.environ[k] = v
ext_modules = [Extension("EvoDAG.sparse_array",
                         [join("EvoDAG", "sparse_array.pxd"),
                          join("EvoDAG", "sparse_array.pyx")],
                         # libraries=["m"],
                         include_dirs=[numpy.get_include()])]

with open('README.rst') as fpt:
    long_desc = fpt.read()
version = open("VERSION").readline().lstrip().rstrip()
lst = open(join("EvoDAG", "__init__.py")).readlines()
for k in range(len(lst)):
    v = lst[k]
    if v.count("__version__"):
        lst[k] = "__version__ = '%s'\n" % version
with open(join("EvoDAG", "__init__.py"), "w") as fpt:
    fpt.write("".join(lst))

setup(
    name="EvoDAG",
    description="""Evolving Directed Acyclic Graph""",
    long_description=long_desc,
    version=version,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],
    url='https://github.com/mgraffg/EvoDAG',
    author="Mario Graff",
    author_email="mgraffg@ieee.org",
    cmdclass={"build_ext": build_ext, "clean": Clean},
    ext_modules=ext_modules,
    packages=['EvoDAG', 'EvoDAG/tests'],
    include_package_data=True,
    zip_safe=False,
    package_data={'': ['*.pxd']},
    install_requires=['cython >= 0.19.2', 'numpy >= 1.6.2'],
    entry_points={
        'console_scripts': ['EvoDAG=EvoDAG.command_line:main',
                            'EvoDAG-params=EvoDAG.command_line:params',
                            'EvoDAG-train=EvoDAG.command_line:train',
                            'EvoDAG-predict=EvoDAG.command_line:predict'],
    }
)



# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from test_root import cl, X
from EvoDAG.model import Model
from EvoDAG.node import Variable, Add, Mul, Sin
from EvoDAG.gp import Individual
from nose.tools import assert_almost_equals
import numpy as np


def test_indindividual_decision_function():
    Add.nargs = 2
    Mul.nargs = 2
    vars = Model.convert_features(X)
    for x in vars:
        x._eval_ts = x._eval_tr.copy()
    vars = [Variable(k, weight=1) for k in range(len(vars))]
    for i in range(len(vars)):
        ind = Individual([vars[i]])
        ind.decision_function(X)
        hy = ind._ind[0].hy.tonparray()
        [assert_almost_equals(a, b) for a, b in zip(X[:, i], hy)]
    ind = Individual([Sin(0, weight=1),
                      Add(range(2), np.ones(2)), vars[0], vars[-1]])
    ind.decision_function(X)
    hy = ind._ind[0].hy.tonparray()
    y = np.sin(X[:, 0] + X[:, -1])
    [assert_almost_equals(a, b) for a, b in zip(y, hy)]
    y = np.sin((X[:, 0] + X[:, 1]) * X[:, 0] + X[:, 2])
    ind = Individual([Sin(0, weight=1), Add(range(2), weight=np.ones(2)),
                      Mul(range(2), weight=1),
                      Add(range(2), weight=np.ones(2)),
                      vars[0], vars[1], vars[0], vars[2]])
    ind.decision_function(X)
    # assert v.hy.SSE(v.hy_test) == 0
    hy = ind._ind[0].hy.tonparray()
    [assert_almost_equals(a, b) for a, b in zip(hy, y)]
    

def test_gp_population_full():
    Add.nargs = 2
    Mul.nargs = 2
    from EvoDAG.gp import Population
    from EvoDAG import EvoDAG
    fs = EvoDAG()._function_set

    class Population2(Population):
        def __init__(self, *args, **kwargs):
            super(Population2, self).__init__(*args, **kwargs)
            self._funcs = [Add, Sin]
            self._terms = [2, 0]

        def random_function(self):
            func = self._funcs.pop()
            if func.nargs == 1:
                return func(0, weight=1)
            return func(range(func.nargs), weight=np.ones(func.nargs))

        def random_terminal(self):
            return Variable(self._terms.pop(), 1)

    pop = Population2(fs, nterminals=3)
    ind = pop.create_random_ind_full(depth=2)
    assert len(pop._funcs) == 0 and len(pop._terms) == 0
    assert isinstance(ind[0], Sin) and isinstance(ind[1], Add)
    assert ind[2].variable == 0 and ind[3].variable == 2
    ind = Individual(ind)
    print(X.shape, ind.individual)
    hy = ind.decision_function(X)
    assert hy.isfinite()


def test_gp_population_grow():
    Add.nargs = 2
    Mul.nargs = 2
    from EvoDAG.gp import Population
    from EvoDAG import EvoDAG
    fs = EvoDAG()._function_set

    class Population2(Population):
        def __init__(self, *args, **kwargs):
            super(Population2, self).__init__(*args, **kwargs)
            self._funcs = [Sin, Add]
            self._terms = [2, 0]
            self._g_funcs = [0, 1, 1]

        def random_function(self):
            func = self._funcs.pop()
            if func.nargs == 1:
                return func(0, weight=1)
            return func(range(func.nargs), weight=np.ones(func.nargs))

        def random_terminal(self):
            return Variable(self._terms.pop(), 1)

        def grow_use_function(self, depth=0):
            if depth == 0:
                return False
            return self._g_funcs.pop()

    pop = Population2(fs, nterminals=3)
    ind = pop.create_random_ind_grow(depth=2)
    print(ind)
    assert len(pop._funcs) == 0 and len(pop._terms) == 0
    assert len(pop._g_funcs) == 0
    assert isinstance(ind[0], Add) and isinstance(ind[1], Sin)
    assert ind[2].variable == 0 and ind[3].variable == 2
    ind = Individual(ind)
    print(X.shape, ind.individual)
    hy = ind.decision_function(X)
    assert hy.isfinite()


def test_gp_create_population():
    Add.nargs = 2
    Mul.nargs = 2
    from EvoDAG.gp import Population
    from EvoDAG import EvoDAG
    fs = EvoDAG()._function_set
    pop = Population(fs, nterminals=X.shape[1])
    inds = pop.create_population(X=X)
    for i in inds:
        assert i[0].hy.isfinite()
